from mlagents_envs.environment import ActionTuple, UnityEnvironment
import EnvChannel
import TaskSideChannel
import PickupAndDeliveryChannel as PADEC
import numpy as np
import os
import time

from Node import Node

import configparser

from lkh import LKH

# from pickup_and_delivery.task import Task
from ta_system import System as TA_System
from ta_state import Location
from ta_agent import Agent

# as long as simulation_going is true, new actions will be generated
simulation_going = True

# Load the side channels and Unity environment.
envChannel = EnvChannel.EnvChannel()
taskChannel = TaskSideChannel.TaskSideChannel()
# Open unity and press play to execute
env = UnityEnvironment(file_name=None, side_channels=[
                       envChannel, taskChannel])
env.reset() # reset the environment

# Specify the used behaviour name for steering the agent
behavior_names = list(env.behavior_specs.keys())
behavior_name = behavior_names[0]

# info obtained from environment
environment_info = envChannel.EnvData
# set up links between task side channel and environment info
environment_info.task_side_channel = taskChannel
taskChannel.environment_data = environment_info

##--------##
## Agents ##
##--------##

# Create all the agents
agentsList = []
for robot_id, start_node in environment_info.current_nodes_dictionary.items():
    # Convert the Node that represents the agent's starting position to a Location
    robot_position = Location(start_node.y, start_node.x, start_node.g)
    # Create agent
    agentsList.append(Agent(robot_id, robot_position))

##-------##
## Tasks ##
##-------##

# Retrieve all tasks
tasksList = environment_info.tasks_to_be_assigned

##-----------------------------##
## LKH3-Solver: Task sequences ##
##-----------------------------##

## Create the LKH class so that the problem (and parameter) files for the LKH3 solver can be generated
lkh = LKH(environment_info.matrix, tasksList, agentsList)
# File that contains all the task sequences: can either be pre-computed or can be computed on the fly
lkhFile = lkh.get_task_sequences()

##--------##
## System ##
##--------##

# Initialize system
system = TA_System(environment_info.matrix, environment_info.rotation, agentsList, tasksList, lkhFile)

##---------------##
## Path Planning ##
##---------------##

# Retrieve all agents from the system
agents_list = system.world.agents
# Retrieve all paths from the system
agent_paths = system.get_paths()

##-------------##
## Follow Path ##
##-------------##

# Keep track of current timestep
actions_requested_in_timestep = 0
timestep = 0

# Keep track of tasks that need to be assigned in Unity
to_assign_tasks = []

# Assign all agents to their first task in Unity
for a_id, agent in agents_list.items():
    # Agent has at least one task
    if len(agent.taskSet) > 0:
        # Get the agent's first task
        next_task = agent.taskSet[0]
        # Assign the first task to the agent in Unity
        environment_info.assign_task(
            a_id, system.tasks[next_task.id])

while simulation_going:    
    # Get information on the agents that request an action
    decision_steps, terminal_steps = env.get_steps(behavior_name)
    agents = decision_steps.agent_id
    observation = decision_steps.obs[0][0]

    # Actions that will be sent to Unity
    actions = []
    action_tuple = ActionTuple()  # actions to send to unity

    # Assign tasks in Unity for agents that are waiting for a new task assignment
    for agent_tuple in to_assign_tasks:
        # Extract the needed information from the tupple
        agent_id = agent_tuple[0]
        task = agent_tuple[1]
        # Assign the task to the agent in Unity
        environment_info.assign_task(agent_id, task)
        # Remove the task from the tupple list to avoid assigning the same task to the same agent more than once
        to_assign_tasks.remove(agent_tuple)

    ##--------------##
    ## Observations ##
    ##--------------##

    # get observations per robot and store in dictionary observations[instance_id] -> observation
    observations = {}
    for i in range(environment_info.number_of_robots):
        observation_i = observation[i * environment_info.observation_size:
                                    i * environment_info.observation_size + environment_info.observation_size]
        observations[observation_i[0]] = observation_i[1:]
    
    # Create an action for all agents that request an action
    for robot_id_unity in observations:   
        
        ##--------------##
        ## OBSERVATIONS ##
        ##--------------##
        
        # robot_id_unity = list(agents_list.keys())[i]
        # Collect agent's observations        
        observation = observations[robot_id_unity]

        # Extract observations
        # robot_id_unity = observation[0] # robot id in unity used to get correct precomputed path
        robot_pos_x = observation[1]
        robot_pos_y = observation[0]
        robot_rot = observation[2]
        # Convert coordinates to position
        robot_position = Location(robot_pos_x, robot_pos_y, robot_rot)
        
        ##-------------##
        ## FOLLOW PATH ##
        ##-------------##

        # Get the pre-computed path of the agent
        path = agent_paths[robot_id_unity]

        next_pos = None

        # Check if agent still needs to move
        if timestep <= agents_list[robot_id_unity].final_action:
            # Agent has not yet reached the end of its path: get the next position of the agent
            next_pos = path[timestep]
        else:
            # Agent has reached the end of its path: stay parked at its parking location
            next_pos = Location(robot_pos_x, robot_pos_y, robot_rot)

        # Convert Position to Node
        next_action = Node(next_pos.row, next_pos.col, next_pos.rotation)

        # Check whether agent needs to pickup/deliver
        pickup_action = 0
        delivery_action = 0
        if next_pos.to_pickup:
            # Agent has to pick up his task
            pickup_action = 1
        elif next_pos.to_deliver:
            # Agent has to deliver his task
            delivery_action = 1
            # Remove the task that was delivered from the taskSet
            agents_list[robot_id_unity].taskSet = agents_list[robot_id_unity].taskSet[1:]
            # Get the task set of the agent
            tasks = agents_list[robot_id_unity].taskSet
            # If there are remaining tasks assign the next one to the agent:
            if (len(tasks) > 0):
                # Get the agent's next task
                next_task = tasks[0]
                # Do not assign the task directly as the agent has not yet delivered it's task, add it to a list so that it will be assigned in the next timestep
                to_assign_tasks.append((robot_id_unity, system.tasks[next_task.id]))

        # Send action to agent in Unity
        actions.extend([robot_id_unity, next_action.x, next_action.y, next_action.g, pickup_action, delivery_action])

    timestep += 1
            
    # Send actions to Unity
    actions = np.array(actions)
    actions.resize(1, environment_info.number_of_robots * environment_info.action_size)
    action_tuple.add_discrete(actions)

    env.set_actions(behavior_name, action_tuple)
    env.step()


