import numpy as np
import configparser
import os
import time
import sys
import json

from ta_task import Task
# from position import Position, dist
from ta_world import Location

def dist(loc1, loc2):
    return abs(loc1.col - loc2.col) + abs(loc1.row - loc2.row)

##-----------##
## CONSTANTS ##
##-----------##

# Dummy agents only used to calculate the distances in the graph
class Agent:
    def __init__(self, pos):
        # Dummy agents only have a position
        self.parking_location = pos

##--------##
## System ##
##--------##

class LKH:
    def __init__(self, environment, tasks, agentsList):
        # Keep track of the agents and the amount of agents
        self.agents = agentsList

        # Keep track of all endpoints in the environment
        self.task_endpoints = dict()
        # Initialize environment
        self.init_env(environment)

        # Keep track of tasks in the task set
        self.tasks = tasks
        # Initialize tasks
        self.load_tasks()

    def init_env(self, env):
        # Keep track of task endpoints
        n_endpoints = 0
        for y in range(len(env)):
            for x in range(len(env[0])):
                if env[y][x] == 3:
                    self.task_endpoints[(x, y)] = n_endpoints
                    n_endpoints += 1

    ##-------##
    ## TASKS ##
    ##-------##

    def load_tasks(self):
        for task in self.tasks:
            pickup_id = self.task_endpoints[(
                task.pickup_endpoint.col, task.pickup_endpoint.row)]
            task.pickup_id = pickup_id
            delivery_id = self.task_endpoints[(
                task.delivery_endpoint.col, task.delivery_endpoint.row)]
            task.delivery_id = delivery_id

    ##-------##
    ## LKH-3 ##
    ##-------##
    
    def get_task_sequences(self):

        ##----------------##
        ## Initialization ##
        ##----------------##

        #print(f"Current directory: {os.getcwd()}")

        # Create a configParser to read the config file
        config = configparser.ConfigParser()
        # Read the config file
        config.read("lkh.conf")

        # Extract information needed from config file
        # Path to pre-computed .tour file OR None (default: None)
        tour_file_path = config["lkh"]["tourFile"]
        lkh_path = config["lkh"]["path"]  # Path to the LKH solver (default: /LKH3/LKH)
        # Maximum time (in seconds) that is going to be used by the LKH3 solver to compute task sequences (default: 500)
        lkh_time = config["lkh"]["lkhMaxTime"]

        if tour_file_path != "None":
            # Use a pre-computed .tour file (contains pre-computed task sequences)
            return tour_file_path
        
        # First compute the task sequences using the LKH3 solver, then return the created .tour file

        ## Generate a unique filename for the generated problem, parameter and .tour files
        timestr = time.strftime("%Y_%m_%d-%H_%M_%S")
        lkh_filename = "LKH_" + timestr

        ## Create problem and parameter files
        self.generate_problem_files(lkh_filename)

        ##------------##
        ## Run solver ##
        ##------------##

        ## Run the LKH-3 algorithm to generate .tour file
        #print("\n\n[LKH3] Generating task sequences:\n")
        os.system(lkh_path + " " + "LKH3/PROBLEM/" + lkh_filename + ".par")
        #print("\n\n[LKH3] Task sequences generated")

        ##---------------##
        ## Return result ##
        ##---------------##

        # Save the path of the newly created .tour file
        lkhFile = "LKH3/TOUR/" + lkh_filename + ".tour"
        return lkhFile

        

    def generate_problem_files(self, lkhFileName):

        ##-------------------------------------##
        ## Create graph to calculate sequences ##
        ##-------------------------------------##

        n_tasks = len(self.tasks)
        n_agents = len(self.agents)
        n_nodes = n_tasks + n_agents
        edge_weights = np.zeros((n_nodes, n_nodes))

        # Edge
        for k in range(n_agents):
            for i in range(n_tasks):
                task = self.tasks[i]
                agent = self.agents[k]
                # Edges from tasks to agents
                edge_weights[n_agents +
                             i][k] = dist(task.pickup_endpoint, task.delivery_endpoint)
                # Edges from agents to tasks
                edge_weights[k][n_agents+i] = max(
                    dist(agent.parking_location, task.pickup_endpoint), task.release_time)

        for i in range(n_tasks):
            for j in range(n_tasks):
                if (i != j):
                    taski = self.tasks[i]
                    taskj = self.tasks[j]
                    # Edges between tasks
                    d = dist(taski.pickup_endpoint, taski.delivery_endpoint) + \
                        dist(taski.delivery_endpoint, taskj.pickup_endpoint)
                    edge_weights[n_agents+i][n_agents+j] = d

        #---------------------#
        # Create problem file #
        #---------------------#

        problemFileName = "LKH3/PROBLEM/" + lkhFileName + ".tsp"
        problemFile = open(problemFileName, "w")
        problemFile.write("NAME: MAPD\n")
        problemFile.write("TYPE: TSPTW\n")
        problemFile.write(f"DIMENSION: {n_nodes}\n")
        problemFile.write("EDGE_WEIGHT_TYPE: EXPLICIT\n")
        problemFile.write("DISPLAY_DATA_TYPE: NO_DISPLAY\n")
        problemFile.write("EDGE_WEIGHT_FORMAT: FULL_MATRIX\n")
        problemFile.write("EDGE_WEIGHT_SECTION\n")
        for i in range(n_nodes):
            for j in range(n_nodes):
                problemFile.write(f"{str(int(edge_weights[i][j]))} ")
            problemFile.write("\n")
        problemFile.write("TIME_WINDOW_SECTION\n")
        for i in range(n_agents):
            problemFile.write(f"{i+1} -1 {10000} 0 0 0\n")
        for i in range(n_tasks):
            task = self.tasks[i]
            problemFile.write(
                f"{i+1+n_agents} {task.release_time} {10000} {dist(task.pickup_endpoint, task.delivery_endpoint)} {task.pickup_id} {task.delivery_id}\n")

        #-----------------------#
        # Create parameter file #
        #-----------------------#

        parameterFileName = "LKH3/PROBLEM/" + lkhFileName + ".par"
        parFile = open(parameterFileName, "w")
        parFile.write("SPECIAL\n")
        parFile.write("PROBLEM_FILE = " + problemFileName + "\n")
        parFile.write("MAX_TRIALS = 100000\nTIME_LIMIT = 1000\n")
        parFile.write("RUNS = 1\n")
        parFile.write("TRACE_LEVEL = 1\n")
        parFile.write("TOUR_FILE = LKH3/TOUR/" + lkhFileName + ".tour\n")
        parFile.write("MAKESPAN = YES\n")






