"""
Created on Wed Apr 22 07:43:03 2020

@author: Pieter Cawood

"""
from heapq import *
from ta_state import *
from ta_world import MAPNODETYPES

def rotation_steps(rotation, start_rotation, end_rotation):
    n_rotations = int(360/rotation)
    return int(abs(end_rotation-start_rotation)/rotation)
    


def get_heuristic(start_location, goal_location, rotation):
    # Admissible heuristic - manhattan distance
    return abs(start_location.col - goal_location.col) + abs(start_location.row - goal_location.row) + rotation_steps(rotation, start_location.rotation, goal_location.rotation)

def get_old_heuristic(start_location, goal_location):
    # Admissible heuristic - manhattan distance
    return abs(start_location.col - goal_location.col) + abs(start_location.row - goal_location.row)


class AStar:
    def __init__(self, world, _rotation):
        self.dim_x = world.width
        self.dim_y = world.height
        self.obstacles = {}
        self.other_agent_paths = []

        self.rotation = _rotation

    def state_possible(self, current_state, new_state):
        # Check not parking or wall
        not_obstacle = False
        if (new_state.location.col, new_state.location.row) not in self.obstacles:
            not_obstacle = True
        return 0 <= new_state.location.col < self.dim_x and 0 <= new_state.location.row < self.dim_y \
               and not_obstacle and not self.is_collision(current_state, new_state)

    def is_collision(self, current_state, new_state): 
        for other_agent_path in self.other_agent_paths:
            # Vertex collision
            if new_state.time in other_agent_path:
                if other_agent_path[new_state.time].row == new_state.location.row and other_agent_path[new_state.time].col == new_state.location.col:
                    return True
            # Edge collision
            if new_state.time in other_agent_path and current_state.time in other_agent_path:
                if other_agent_path[new_state.time].row == current_state.location.row and \
                    other_agent_path[new_state.time].col == current_state.location.col and \
                        other_agent_path[current_state.time].row == new_state.location.row and \
                            other_agent_path[current_state.time].col == new_state.location.col:
                    return True
        return False

    def get_neighbours(self, current_state):
        neighbours = []
        neighbour_time_step = current_state.time + 1
        # Wait in state
        new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row, current_state.location.rotation))
        if self.state_possible(current_state, new_state):
            neighbours.append(new_state)

        # Add rotation
        new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row, (current_state.location.rotation+self.rotation) % 360))
        if self.state_possible(current_state, new_state):
            neighbours.append(new_state)
        # Remove rotation
        new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row, (current_state.location.rotation-self.rotation) % 360))
        if self.state_possible(current_state, new_state):
            neighbours.append(new_state)

        ##  Move in specific directions
        if current_state.location.rotation == 0:
            # Move down
            new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row - 1, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)        
        elif current_state.location.rotation == 90:
            # Move right
            new_state = State(neighbour_time_step, Location(current_state.location.col + 1, current_state.location.row, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)
        elif current_state.location.rotation == 180:
            # Move up
            new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row + 1, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)
        elif current_state.location.rotation == 270:
            new_state = State(neighbour_time_step, Location(current_state.location.col - 1, current_state.location.row, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)
                 
        return neighbours

    def create_obstacles(self, world, other_agent_paths, remaining_parking, timestep):
        # Map blocked cells
        for element in world:
            if world[element] == MAPNODETYPES.WALL:
                self.obstacles[(element[0], element[1])] = True
                    
        # Other agents' parking locations
        for location in remaining_parking:
            self.obstacles[(location.col, location.row)] = True

        short_paths = []
        t = timestep
        for path in other_agent_paths:
            new_path = dict()
            t = timestep
            while t in path:
                new_path[t] = path[t]
                t += 1
            short_paths.append(new_path)
        self.other_agent_paths = short_paths

    def search(self, time_step, start_location, goal_location, other_agent_paths, world, parking_locations, agent,
               release_time=0, find_step = 0):
        self.create_obstacles(world, other_agent_paths, parking_locations, time_step)
        initial_state = State(time_step, start_location)
        came_from = {}
        open_list = [initial_state]
        closed_set = set()
        initial_state.g_score = 0
        # initial_state.f_score = get_heuristic(start_location, goal_location, self.rotation)
        initial_state.f_score = get_old_heuristic(start_location, goal_location)
        while open_list:
            current = heappop(open_list)
            # If agent is at the goal location
            if current.location == goal_location:
                # Agent cannot move during pickup/delivery, needs at least 2 timesteps at the goal location
                if current in came_from and came_from[current].location == goal_location:
                    # If not yet release time hold and run search
                    if not current.time < release_time:
                        new_path = [current]
                        while current in came_from:
                            current = came_from[current]
                            new_path.append(current)
                        return new_path[::-1]
            # Continue looking for the goal location
            closed_set |= {current}
            neighbours = self.get_neighbours(current)
            neighbour: State
            for neighbour in neighbours:
                # Only explore new states
                if (neighbour not in closed_set) and (neighbour not in open_list):
                    came_from[neighbour] = current
                    neighbour.g = current.g + 1
                    # neighbour.f = neighbour.g + get_heuristic(neighbour.location, goal_location, self.rotation)
                    neighbour.f = neighbour.g + get_old_heuristic(neighbour.location, goal_location)
                    heappush(open_list, neighbour)
        return False
