"""
Created on Sat Apr 18 11:11:10 2020

@author: Pieter Cawood
"""
import time
from ta_world import *
from ta_task import *
from ta_astar import AStar
from queue import Queue
from itertools import combinations
import concurrent.futures

GLOBAL_MAX_AGENT_TIME = 2500

def plan_path(world, rotation, agent, time_step, release_time=0, find_step=0):
        a_star = AStar(world, rotation)
        other_agent_paths = []
        parking_locations = []

        # Add other agent paths
        for agent_id in world.agents:
            if agent_id != agent.id:
                other_agent_paths.append(world.agents[agent_id].path)
                parking_locations.append(world.agents[agent_id].parking_location)
        start_location = agent.path[time_step]
        goal_location = agent.next_endpoint

        # Non dummy path, add release time to hold pickup
        new_path = a_star.search(time_step, start_location, goal_location, other_agent_paths, world, parking_locations,
                                agent, release_time, find_step)
        # Could not find path
        if not new_path:
            return False

        # Update agent path
        for state in new_path:
            world.agents[agent.id].path[state.time] = state.location

        world.agents[agent.id].time_at_goal = time_step + len(new_path) - 1
        return world.agents[agent.id].time_at_goal

def copy_queue(queue):
        result = Queue()
        for entry in queue.queue:
            result.put(entry)
        return result

def calculate_execution_time(world, rotation, agent_id, tsp_seqs):
        # print(f"Calculating execution time of agent: {agent_id}")
        time_step = 0
        new_start_step = -1
        while True:
            new_start_step += 1
            time_step = new_start_step
            if agent_id not in tsp_seqs:
                tsp_seqs[agent_id] = Queue()
            seq = copy_queue(tsp_seqs[agent_id])
            path_failed = False
            for seq_task in range(seq.qsize()):
                task = seq.get()
                # To pickup position
                world.agents[agent_id].next_endpoint = task.pickup_endpoint
                time_step = plan_path(world, rotation, world.agents[agent_id], time_step, task.release_time, 1)
                if not time_step:
                    path_failed = True
                    break
                # To delivery position
                world.agents[agent_id].next_endpoint = task.delivery_endpoint
                time_step = plan_path(world, rotation, world.agents[agent_id], time_step, 0, 2)
                if not time_step:
                    path_failed = True
                    break
                if not path_failed:
                    break
            # Move agent path back to parking position
            for t_step in range(GLOBAL_MAX_AGENT_TIME):
                world.agents[agent_id].path[t_step] = world.agents[agent_id].parking_location
            return time_step

class System:
    def __init__(self, env, rotation, agentsList, taskList, sequence_file_path):
        self.env = env
        self.rotation = rotation
        self.agentsList = agentsList
        self.taskList = taskList
        self.sequence_file_path = sequence_file_path

        self.world = None
        self.tasks = None

        self.paths = dict()

        self.makespan = 0

        self.main()


    def load_files(self):
        world = World(self.env, self.agentsList)
        self.world = world
        tasks = TaskDict(self.taskList, world)
        self.tasks = tasks
        agent_tour_dict = TourDict(
            self.sequence_file_path, self.agentsList, tasks)
        return world, tasks, agent_tour_dict


    def find_agent_path(self, world, rotation, agent, time_step, release_time=0, find_step=0):
        a_star = AStar(world, rotation)
        other_agent_paths = []
        parking_locations = []

        # Add other agent paths
        for agent_id in world.agents:
            if agent_id != agent.id:
                other_agent_paths.append(world.agents[agent_id].path)
                parking_locations.append(world.agents[agent_id].parking_location)
        start_location = agent.path[time_step]
        goal_location = agent.next_endpoint

        # Non dummy path, add release time to hold pickup
        new_path = a_star.search(time_step, start_location, goal_location, other_agent_paths, world, parking_locations,
                                agent, release_time, find_step)
        # Could not find path
        if not new_path:
            return False

        # Update agent path
        for state in new_path:
            world.agents[agent.id].path[state.time] = state.location

        world.agents[agent.id].time_at_goal = time_step + len(new_path) - 1
        return world.agents[agent.id].time_at_goal


    def test_paths(self, world):
        for agent_1, agent_2 in combinations(world.agents, 2):
            for time_step in world.agents[agent_1].path:
                if time_step > max(world.agents[agent_2].path):
                    break
                # Vertex collision between any two agents
                if world.agents[agent_1].path[time_step] == world.agents[agent_2].path[time_step]:
                    #print("Agent #" + str(agent_1) + ",#" + str(agent_2) + " vertex collision at " +
                        #str(world.agents[agent_1].path[time_step]) + " t:" + str(time_step))
                    return False
                # Edge collision between any two agents
                elif (time_step + 1 in world.agents[agent_1].path) and (time_step + 1 in world.agents[agent_2].path):
                    if (world.agents[agent_1].path[time_step] == world.agents[agent_2].path[time_step + 1]) and \
                            (world.agents[agent_2].path[time_step] == world.agents[agent_1].path[time_step + 1]):
                        #print("Agent #" + str(agent_1) + ",#" + str(agent_2) + " edge collision at " +
                            #str(world.agents[agent_1].path[time_step]) + "-" +
                            #str(world.agents[agent_2].path[time_step]) + " t:" + str(time_step))
                        return False
        return True



    def main(self):
        world, tasks, tsp_seqs = self.load_files()
        make_span = 0
        last_sim_step = 0
        closed_list = {}
        time_start = time.perf_counter()
        for next_agent in world.agents:
            time_agent_start = time.perf_counter()
            max_len = 0
            priority_id = -1
            #print(f"Determining execution times for remaining agents: ", end='')
            '''
            Calculate agents execution time for their paths
            Initialise path to parking position once complete
            '''
            estimated_times = []
            threads = []
            remainingAgents = []

            for agent_id in world.agents:
                if agent_id not in closed_list:
                    remainingAgents.append(agent_id)

            if len(remainingAgents) > 1:
                with concurrent.futures.ThreadPoolExecutor() as executor:
                    for agent_id in remainingAgents:
                        #print(agent_id, end=',')
                        future = executor.submit(calculate_execution_time, world, self.rotation, agent_id, tsp_seqs)
                        threads.append((future, agent_id))
                #print("")

                for threadTuple in threads:
                    timestep = threadTuple[0].result()
                    estimated_times.append((timestep, threadTuple[1]))
                largest_exec = max(estimated_times, key=lambda t: t[0])
                max_len = largest_exec[0]
                priority_id = largest_exec[1] 
            else:
                priority_id = remainingAgents[0]                  
            # print(f"Calculating actual for {priority_id}")
            # Expand _id: highest execution time
            closed_list[priority_id] = True
            # print(f"Agent {priority_id} is now in closed list")
            '''
            Calculate actual prioritized agent path
            '''
            new_start_step = -1
            while True:
                new_start_step += 1
                time_step = new_start_step
                seq = copy_queue(tsp_seqs[priority_id])
                path_failed = False
                for next_task in range(tsp_seqs[priority_id].qsize()):
                    task = seq.get()
                    # To pickup position
                    world.agents[priority_id].next_endpoint = task.pickup_endpoint
                    time_step = self.find_agent_path(world, self.rotation, world.agents[priority_id], time_step, task.release_time, 3)
                    world.agents[priority_id].path[time_step].task_id = task.id
                    world.agents[priority_id].path[time_step].to_pickup = True
                    if not time_step:
                        path_failed = True
                        break
                    # To delivery position
                    world.agents[priority_id].next_endpoint = task.delivery_endpoint
                    time_step = self.find_agent_path(world, self.rotation, world.agents[priority_id], time_step, 0, 4)
                    world.agents[priority_id].path[time_step].task_id = task.id
                    world.agents[priority_id].path[time_step].to_deliver = True
                    if not time_step:
                        path_failed = True
                        break
                    task.delivered = time_step
                    world.agents[priority_id].taskSet.append(task)
                if not path_failed:
                    break
            # Record make_span before parking travel
            make_span = max(make_span, world.agents[priority_id].time_at_goal)
            time_agent_stop = time.perf_counter()
            #print("\t* Agent #" + str(priority_id) + " execution time: " + str(
            #world.agents[priority_id].time_at_goal) + " (calc :" + str(
            #round(((time_agent_stop - time_agent_start) / 60), 1)) + " mins)")
            # End - Travel back to parking
            world.agents[priority_id].next_endpoint = world.agents[priority_id].parking_location
            self.find_agent_path(world, self.rotation, world.agents[priority_id], time_step)
            last_sim_step = max(last_sim_step, world.agents[priority_id].time_at_goal)
            world.agents[priority_id].final_action = last_sim_step

        time_stop = time.perf_counter()
        #print("Make_span of :" + str(make_span))
        self.makespan = make_span
        #print("Runtime of :" + str(round(((time_stop - time_start) / 60), 1)) + " minutes")
        for a_id, a in world.agents.items():
            self.paths[a_id] = a.path

    def get_paths(self):
        return self.paths

    def get_parking_locations(self):
        agents = self.world.agents
        parking_locations = dict()
        for a_id, agent in agents.items():
            parking_pos = agent.parking_location
            parking_locations[parking_pos] = a_id
        return parking_locations
